/*****************************************************************
*
* This is an implementation of decimation-in-time FFT algorithm for
* real sequences.  The techniques used here can be found in several
* books, e.g., i) Proakis and Manolakis, "Digital Signal Processing",
* 2nd Edition, Chapter 9, and ii) W.H. Press et. al., "Numerical
* Recipes in C", 2nd Ediiton, Chapter 12.
*
* Input -  There are two inputs to this function:
*
*       1) An integer pointer to the input data array 
*       2) An integer value which should be set as +1 for FFT
*          and some other value, e.g., -1 for IFFT
*
* Output - There is no return value.
*       The input data are replaced with transformed data.  If the
*       input is a real time domain sequence, it is replaced with
*       the complex FFT for positive frequencies.  The FFT value 
*       for DC and the foldover frequency are combined to form the
*       first complex number in the array.  The remaining complex
*       numbers correspond to increasing frequencies.  If the input
*       is a complex frequency domain sequence arranged as above,
*       it is replaced with the corresponding time domain sequence. 
*
* Notes:
*
*       1) This function is designed to be a part of a noise supp-
*          ression algorithm that requires 128-point FFT of real
*          sequences.  This is achieved here through a 64-point
*          complex FFT.  Consequently, the FFT size information is
*          not transmitted explicitly.  However, some flexibility
*          is provided in the function to change the size of the 
*          FFT by specifying the size information through "define"
*          statements.
*
*       2) The values of the complex sinusoids used in the FFT 
*          algorithm are computed once (i.e., the first time the
*          r_fft function is called) and stored in a table. To
*          further speed up the algorithm, these values can be
*          precomputed and stored in a ROM table in actual DSP
*          based implementations.
*
*       3) In the c_fft function, the FFT values are divided by
*          2 after each stage of computation thus dividing the
*          final FFT values by 64.  No multiplying factor is used
*          for the IFFT.  This is somewhat different from the usual
*          definition of FFT where the factor 1/N, i.e., 1/64, is
*          used for the IFFT and not the FFT.  No factor is used in
*          the r_fft function.
*
*       4) Much of the code for the FFT and IFFT parts in r_fft
*          and c_fft functions are similar and can be combined.
*          They are, however, kept separate here to speed up the
*          execution.
* 
* Modifications: (Tommy Vaillancourt mars 2002)
*          1) Scaling phs_tbl by two to prevent overflow
*          2) Add Switch and case to use r_fft_fx between 32 and 128 fft SIZE      
*          3) Use of L_msu and L_mac rather than L_sub and L_add 
*             in core of c_fft_fx and r_fft_fx (gain of 0.13 MIPS with fft9=9x128)
* Revised and updated by B.Bessette    july 2002
*
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>


#include "typedef.h"
#include "basic_op.h"
#include "count.h"
#include "amr_plus_fx.h"


#define TRUE         1
#define FALSE        0


/* FFT/IFFT function for complex sequences */

/*
 * The decimation-in-time complex FFT/IFFT is implemented below.
 * The input complex numbers are presented as real part followed by
 * imaginary part for each sample.  The counters are therefore
 * incremented by two to access the complex valued samples.
 */

static void c_fft_fx(
  Word16 *farray_ptr,  /* coeffients in the order re[0], re[n/2], re[1], im[1], ..., re[n/2-1], im[n/2-1] */
  Word16 wSize,
  Word16 wSizeByTwo,
  Word16 wNumberStage,
  Word16 *ii_table,
  Word16 *phs_tbl,
  Word16 isign
)                      /* 1=fft, otherwize it's ifft                 */ 
{
  Word16 i, j, k, ii, jj, kk, ji, kj;
  Word32 ftmp, ftmp_real, ftmp_imag;
  Word16 tmp1, tmp2, wSize2;

  wSize2 = sub(wSize, 2);
  /* Rearrange the input array in bit reversed order */
  for (i = 0, j = 0; i < wSize2; i = i + 2)
  {
    test();
    if (j > i)
    {
      ftmp = *(farray_ptr + i);                        move16();
      *(farray_ptr + i) = *(farray_ptr + j);           move16();
      *(farray_ptr + j) = (Word16)ftmp;                move16();

      ftmp = *(farray_ptr + i + 1);                    move16();
      *(farray_ptr + i + 1) = *(farray_ptr + j + 1);   move16();
      *(farray_ptr + j + 1) = (Word16)ftmp;            move16();
    }

    k = wSizeByTwo;     move16();
    test();
    while (j >= k)
    {
      test();  
      j = sub(j, k);
      k = shr(k, 1);
    }
    j += k;
  }

  /* The FFT part */
  if (isign == 1)
  {
    for (i = 0; i < wNumberStage; i++)
    {                         /* i is stage counter */
      jj = shl(2, i);         /* FFT size */
      kk = shl(jj, 1);        /* 2 * FFT size */
      ii = ii_table[i];      move16(); /* 2 * number of FFT's */
      for (j = 0; j < jj; j = j + 2)
      {                               /* j is sample counter */
        ji = shl(j, ii);                  /* ji is phase table index */
        for (k = j; k < wSize; k = k + kk)
        {                             /* k is butterfly top */
          kj = add(k, jj);            /* kj is butterfly bottom */

          /* Butterfly computations */

          ftmp_real = L_msu(L_mult(*(farray_ptr + kj), phs_tbl[ji]),
                                 *(farray_ptr + kj + 1), phs_tbl[ji + 1]);
          ftmp_imag = L_mac(L_mult(*(farray_ptr + kj + 1), phs_tbl[ji]),
                                 *(farray_ptr + kj), phs_tbl[ji + 1]);
                                                        
          ftmp = L_mult(*(farray_ptr + k), 16384);
          ftmp = L_sub(ftmp, ftmp_real);
          *(farray_ptr + kj) = round(ftmp);                  move16();

          ftmp = L_mult(*(farray_ptr + k + 1), 16384);
          ftmp = L_sub(ftmp, ftmp_imag);
          *(farray_ptr + kj + 1) = round(ftmp);              move16();

          ftmp = L_mult(*(farray_ptr + k), 16384);
          ftmp = L_add(ftmp, ftmp_real);
          *(farray_ptr + k) = round(ftmp);                   move16();

          ftmp = L_mult(*(farray_ptr + k + 1), 16384);
          ftmp = L_add(ftmp, ftmp_imag);
          *(farray_ptr + k + 1) = round(ftmp);               move16();
        }
      }
    }
  }
  else    /* The IFFT part */
  {
    for (i = 0; i < wNumberStage; i++)
    {                         /* i is stage counter */
      jj = shl(2, i);         /* FFT size */
      kk = shl(jj, 1);        /* 2 * FFT size */
      ii = ii_table[i];       /* 2 * number of FFT's */

      for (j = 0; j < jj; j = j + 2)
      {                               /* j is sample counter */
        /* This can be computed by successive additions of ii to ji, starting from 0
           hence line-count it as a one-line add (still need to increment op count!!) */

        ji = shl(j, ii);    /* ji is phase table index */
       
        for (k = j; k < wSize; k = k + kk)
        {                             /* k is butterfly top */
          kj = add(k, jj);            /* kj is butterfly bottom */
                                      /* Butterfly computations */
          ftmp_real = L_mac(L_mult(*(farray_ptr + kj), phs_tbl[ji]),
                                 *(farray_ptr + kj + 1), phs_tbl[ji + 1]);
                                       
          ftmp_imag = L_msu(L_mult(*(farray_ptr + kj + 1), phs_tbl[ji]),
                                 *(farray_ptr + kj), phs_tbl[ji + 1]);

          tmp1 = round(L_shl(ftmp_real,1));
          tmp2 = round(L_shl(ftmp_imag,1));

          *(farray_ptr + kj) = sub(*(farray_ptr + k), tmp1);             move16();
          *(farray_ptr + kj + 1) = sub(*(farray_ptr + k + 1), tmp2);     move16();       
          *(farray_ptr + k) = add(*(farray_ptr + k), tmp1);              move16();
          *(farray_ptr + k + 1) = add(*(farray_ptr + k + 1), tmp2);      move16();
        }
      }
    }
  }

  return;
} 


/* initialize fft table : size of fft (size) must be 16, 32, 64 or 128 */
void init_r_fft_fx(
  Word16 size,
  Word16 wSizeByTwo,
  Word16 wNumberStage,
  Word16 *ii_table,
  Word16 *phs_tbl
)
{
  Word16 i, j,*pt_t, i_offset;

  test();test();test();test();  
   move16(); 
  switch(size) {
  case SIZE/8:      /*16*/
    i_offset = 16;
    break;
  case SIZE/4:      /*32*/
    i_offset = 8;
    break;
  case SIZE/2:      /*64*/
    i_offset = 4;
    break;
  case SIZE:       /*128*/
    i_offset = 2;
    break;
  default:
    printf(" invalid fft size!\n");
    exit(0); 
  }

  j = 0; move16();
  /* calculation of the table of phase according to the dimension of the fft */
  for (i = 0; i <size; i += 2)
  {
    phs_tbl[i] = phs_tbl128[j];                move16();
    phs_tbl[i+1] = phs_tbl128[j+1];            move16(); 

    j += i_offset;
  }
  /* calculation of the ii_table according to the dimension of the fft */
  pt_t = ii_table;
  for (i = wNumberStage; i > 0; i--)
  {
      *pt_t++ = i; move16();
  }
  return;
}


/* perform FFT fixed-point for real-valued sequences of length 32, 64 or 128 (see init_r_fft_fx) */
void r_fft_fx(
  Word16 *farray_ptr, /* coeffients in the order re[0], re[1], ... re[n/2], im[1], im[2], ..., im[n/2-1] */
  Word16 wSize,
  Word16 wSizeByTwo,
  Word16 wNumberStage,
  Word16 *ii_table,
  Word16 *phs_tbl,
  Word16 isign
)                     /* 1=fft, otherwize it's ifft                 */ 
{
  Word16 ftmp1_real, ftmp2_real, ftmp2_imag, wsize1;
  Word32 Lftmp1_real, Lftmp1_imag;
  Word16 i, j;
  Word32 Ltmp1;
  Word16 temp[SIZE], *ptIma, *ptReel;    

  /* The FFT part */
  if (isign == 1)
  {
    /* Perform the complex FFT */
    c_fft_fx(farray_ptr, wSize, wSizeByTwo, wNumberStage, ii_table, phs_tbl, isign);

    /* First, handle the DC and foldover frequencies */
    ftmp1_real = *farray_ptr; move16();
    ftmp2_real = *(farray_ptr + 1); move16();
                
    /*    *farray_ptr = add(ftmp1_real, ftmp2_real);           move16(); */
    
    *farray_ptr = sub(add(ftmp1_real, ftmp2_real),(NUM_STAGE/2));           move16();      /* DC have a small offset */

    *(farray_ptr + 1) = sub(ftmp1_real, ftmp2_real);     move16();

    /* Now, handle the remaining positive frequencies */

    for (i = 2, j = wSize - i; i <= wSizeByTwo; i = i + 2, j = sub(wSize, i))
    {
      Lftmp1_real = L_mult(*(farray_ptr + i), 16384);
      Lftmp1_real = L_mac(Lftmp1_real, *(farray_ptr + j), 16384);
      Lftmp1_imag = L_mult(*(farray_ptr + i + 1), 16384);
      Lftmp1_imag = L_msu(Lftmp1_imag, *(farray_ptr + j + 1), 16384);

      ftmp2_real = add(*(farray_ptr + i + 1), *(farray_ptr + j + 1));
      ftmp2_imag = sub(*(farray_ptr + j), *(farray_ptr + i));

      Ltmp1 = L_msu(L_mult(ftmp2_real, phs_tbl[i]),ftmp2_imag, phs_tbl[i + 1]);

      *(farray_ptr + i) = round(L_add(Ltmp1, Lftmp1_real));           move16();

      Ltmp1 = L_mac(L_mult(ftmp2_imag, phs_tbl[i]),ftmp2_real, phs_tbl[i + 1]);

      *(farray_ptr + i + 1) = round(L_add(Ltmp1, Lftmp1_imag));       move16();

      Ltmp1 = L_mac(L_mult(ftmp2_real, phs_tbl[j]),ftmp2_imag, phs_tbl[j + 1]);

      *(farray_ptr + j) = round(L_add(Ltmp1, Lftmp1_real));           move16();

      Ltmp1 = L_msu(L_mult(ftmp2_imag, phs_tbl[j]),ftmp2_real, phs_tbl[j + 1]);

      *(farray_ptr + j + 1) = round(L_negate(L_add(Ltmp1, Lftmp1_imag)));       move16();
    }

    /* change order of coefficient */
    wsize1 = sub(wSize,1);
    ptIma = &temp[wsize1];           move16();
    ptReel = &temp[1];                move16();
    
    for (i = 2; i<wsize1; i+=2)
    {
      *ptReel++ = *(farray_ptr+i);    move16();
      *ptIma-- = *(farray_ptr+i+1);   move16();
    }
    *ptReel = farray_ptr[1];          move16();
    temp[0] = farray_ptr[0];          move16();

    for (i = 0;i < wSize; i++)
    {
      *(farray_ptr+i) = temp[i];      move16();
    }
  }
  else      /* The IFFT part */
  {
    /* change order of coefficient */
    ptIma = &farray_ptr[wSize-1];move16();
    ptReel = &farray_ptr[1];move16();
    ftmp1_real = *farray_ptr; move16();
    ftmp2_real = farray_ptr[wSizeByTwo];   move16();
                        
    for (i = 2; i<wSize; i+=2)
    {
      temp[i] = *ptReel++;            move16();
      temp[i+1] = *ptIma--;           move16(); 
    }
    temp[0] = ftmp1_real;             move16();
    temp[1] = ftmp2_real;             move16();
                        
    for (i = 0; i<wSize; i++)
    {
      farray_ptr[i] = temp[i];        move16();
    }
                        
    /* First, handle the DC and foldover frequencies */
    ftmp1_real = *farray_ptr;move16();
    ftmp2_real = *(farray_ptr + 1);move16();

    Ltmp1 = L_mult(ftmp1_real, 16384);
    Ltmp1 = L_mac(Ltmp1, ftmp2_real, 16384);
    *farray_ptr = round(Ltmp1);                 move16();

    Ltmp1 = L_mult(ftmp1_real, 16384);
    Ltmp1 = L_msu(Ltmp1, ftmp2_real, 16384);
    *(farray_ptr + 1) = round(Ltmp1);           move16();

    /* Now, handle the remaining positive frequencies */
    for (i = 2, j = wSize - i; i <= wSizeByTwo; i = i + 2, j = sub(wSize, i))
    {
      Lftmp1_real = L_mult(*(farray_ptr + i), 16384);
      Lftmp1_real = L_mac(Lftmp1_real, *(farray_ptr + j), 16384);
      Lftmp1_imag = L_mult(*(farray_ptr + i + 1), 16384);
      Lftmp1_imag = L_msu(Lftmp1_imag, *(farray_ptr + j + 1), 16384);

      ftmp2_real = negate(add(*(farray_ptr + i + 1), *(farray_ptr + j + 1)));
      ftmp2_imag = sub(*(farray_ptr + i), *(farray_ptr + j));

      Ltmp1 = L_mac(L_mult(ftmp2_real, phs_tbl[i]), ftmp2_imag, phs_tbl[i + 1]);

      *(farray_ptr + i) = round(L_add(Ltmp1, Lftmp1_real));           move16();

      Ltmp1 = L_msu(L_mult(ftmp2_imag, phs_tbl[i]),ftmp2_real, phs_tbl[i + 1]);

      *(farray_ptr + i + 1) = round(L_add(Ltmp1, Lftmp1_imag));       move16();

      Ltmp1 = L_msu(L_mult(ftmp2_real, phs_tbl[j]),ftmp2_imag, phs_tbl[j + 1]);

      *(farray_ptr + j) = round(L_add(Ltmp1, Lftmp1_real));           move16();

      Ltmp1 = L_mac(L_mult(ftmp2_imag, phs_tbl[j]),ftmp2_real, phs_tbl[j + 1]);

      *(farray_ptr + j + 1) = round(L_negate(L_add(Ltmp1, Lftmp1_imag)));     move16();
    }

    /* Perform the complex IFFT */
    c_fft_fx(farray_ptr, wSize, wSizeByTwo, wNumberStage, ii_table, phs_tbl, isign);
  }

  return;
}                                                              



