/*-------------------------------------------------------------------*
 * functions: q_isf_hf()                                             *
 *                                                                   *
 * Coding/Decoding of HF ISF parameters with AR prediction.          *
 *                                                                   *
 * The ISF vector is quantized using one-stage VQ with 8 elements.   *
 *-------------------------------------------------------------------*/



#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#include "amr_plus_fx.h"




#define ISF_GAP_HF      180.0f
#define ISF_GAP_HF_FX   461
#define MU_FX           16384

#define MU              0.5f


/*-------------------------------------------------------------------*
 * routine:   d_isf_hf()                                             *
 *            ~~~~~~~~~                                              *
 * Decoding of ISF parameters                                        *
 *-------------------------------------------------------------------*
 *  Arguments:                                                       *
 *    indice[] : indices of the two selected codebook entries        *
 *    isf[]    : quantized ISFs (in frequency domain)                *
 *-------------------------------------------------------------------*/

#define ALPHA_FX     29491

void Q_isf_hf(
  Word16 *isf1,      /* input : ISF in the frequency domain (0..6400) */
  Word16 *isf_q,     /* output: quantized ISF                         */
  Word16 *past_q,    /* i/o   : past quantized isf (for AR prediction)*/
  Word16 *indice,      /* output: quantization indices (7 words)        */
  const Word16 *Mean_isf_hf,
  const Word16 *Dico1_isf_hf
)
{
  Word16 i, k, index, tmp , iQ0, iQ1;
  Word16 isf2[Q_ISF_ORDER];
  Word32 min_err, distance;

  distance = MAX_32;						    move32();
  for (k=0; k<SIZE_BK1_HF; k++)
  {
										
    for (i=0; i<Q_ISF_ORDER; i++) 
    {
      /*isf2[i] = isf1[i] - dico1_isf_hf[(k*Q_ISF_ORDER)+i] - mean_isf_hf[i] - MU*past_q[i]; */
      tmp = sub(isf1[i], Dico1_isf_hf[(k*Q_ISF_ORDER)+i]);
      tmp = sub(tmp, Mean_isf_hf[i]);
      isf2[i] = sub(tmp, mult_r(past_q[i], MU_FX));    move16();
    }

    index = Sub_VQ(isf2, Dico2_isf_hf, Q_ISF_ORDER, SIZE_BK2_HF, &min_err);	
  
    test();
    if (L_sub(min_err,distance) < 0)
    {
      distance = min_err;			  move32();			
      indice[0] = k;            move16();  
      indice[1] = index;        move16();
    }
  }

  iQ0 = shl(indice[0],3);
  iQ1 = shl(indice[1],3); 
									
  for (i=0; i<Q_ISF_ORDER; i++) 
  {
    /*isf_q[i] = dico2_isf_hf[(indice[1]*Q_ISF_ORDER)+i] + dico1_isf_hf[(indice[0]*Q_ISF_ORDER)+i]
               + mean_isf_hf[i] + MU*past_q[i];					*/
      tmp = add(Dico1_isf_hf[iQ0+i],Dico2_isf_hf[iQ1+i]); 
      tmp = add(tmp, Mean_isf_hf[i]);
      isf_q[i] = add(tmp, mult_r(MU_FX, past_q[i]));   move16();
  }

  Reorder_isf(isf_q, ISF_GAP_HF_FX, Q_ISF_ORDER);	
  
  for (i=0; i<Q_ISF_ORDER; i++) 
  {
    past_q[i] = sub(isf_q[i],Mean_isf_hf[i]);					    move16();
  }
	
  return;
}



void D_isf_hf(
  Word16 *indice,      /* (i) : quantization indices                    */
  Word16 *isf_q,       /* (o) : quantized ISFs in the cosine domain     */
  Word16 *past_q,      /* i/o : past quantized isf (for AR prediction)  */
  Word16 bfi,          /* (i) : Bad frame indicator                     */
  const Word16 *Mean_isf_hf, /* (i) : Mean isf Dico                           */
  const Word16 *Dico1_isf_hf /* (i) : sif Dico                                */
)
{
  Word16 i, iQ0, iQ1, tmp;
  

  iQ0 = shl(indice[0],3);
  iQ1 = shl(indice[1],3); 
  
  test();
  if(bfi == 0) /* Good frame */
  {
    for (i = 0; i < Q_ISF_ORDER; i++) 
    {
      tmp = add(Dico1_isf_hf[iQ0+i],Dico2_isf_hf[iQ1+i]); 
      tmp = add(tmp, Mean_isf_hf[i]);
      isf_q[i] = add(tmp, mult_r(MU_FX, past_q[i]));   move16();
    }
  }
  else /* bad frame */
  {
    /* use the past ISFs slightly shifted towards the means */
    for (i=0; i<Q_ISF_ORDER; i++) 
    {
      isf_q[i] = add(mult_r(ALPHA_FX,past_q[i]), Mean_isf_hf[i]);   move16();
    }
  }

  Reorder_isf(isf_q, ISF_GAP_HF_FX, Q_ISF_ORDER);

  for (i=0; i<Q_ISF_ORDER; i++) 
  {
    past_q[i] = sub(isf_q[i] , Mean_isf_hf[i]);    move16();
  }

  return;
}


